@@ -12,9 +12,8 @@
 /*	void	(*service)(VSTREAM *stream, char *service_name, char **argv);
 /*	int	key;
 /*
-/*	void	multi_server_disconnect(stream, argv)
+/*	void	multi_server_disconnect(stream)
 /*	VSTREAM *stream;
-/*	char	**argv;
 /*
 /*	void	multi_server_drain()
 /* DESCRIPTION
@@ -36,53 +35,71 @@
 /*	function is run after the program has optionally dropped its
 /*	privileges. This function should not attempt to preserve state
 /*	across calls. The stream initial state is non-blocking mode.
+/*	Optional connection attributes are provided as a hash that
+/*	is attached as stream context. NOTE: the attributes are
+/*	destroyed after this function is called.
 /*	The service name argument corresponds to the service name in the
 /*	master.cf file.
 /*	The argv argument specifies command-line arguments left over
 /*	after options processing.
 /* .PP
-/*	Optional arguments are specified as a null-terminated (key, value)
-/*	list. Keys and expected values are:
-/* .IP "MAIL_SERVER_INT_TABLE (CONFIG_INT_TABLE *)"
+/*	Optional arguments are specified as a null-terminated list
+/*	with macros that have zero or more arguments:
+/* .IP "CA_MAIL_SERVER_INT_TABLE(CONFIG_INT_TABLE *)"
 /*	A table with configurable parameters, to be loaded from the
 /*	global Postfix configuration file. Tables are loaded in the
 /*	order as specified, and multiple instances of the same type
 /*	are allowed.
-/* .IP "MAIL_SERVER_STR_TABLE (CONFIG_STR_TABLE *)"
+/* .IP "CA_MAIL_SERVER_LONG_TABLE(CONFIG_LONG_TABLE *)"
 /*	A table with configurable parameters, to be loaded from the
 /*	global Postfix configuration file. Tables are loaded in the
 /*	order as specified, and multiple instances of the same type
 /*	are allowed.
-/* .IP "MAIL_SERVER_BOOL_TABLE (CONFIG_BOOL_TABLE *)"
+/* .IP "CA_MAIL_SERVER_STR_TABLE(CONFIG_STR_TABLE *)"
 /*	A table with configurable parameters, to be loaded from the
 /*	global Postfix configuration file. Tables are loaded in the
 /*	order as specified, and multiple instances of the same type
 /*	are allowed.
-/* .IP "MAIL_SERVER_TIME_TABLE (CONFIG_TIME_TABLE *)"
+/* .IP "CA_MAIL_SERVER_BOOL_TABLE(CONFIG_BOOL_TABLE *)"
 /*	A table with configurable parameters, to be loaded from the
 /*	global Postfix configuration file. Tables are loaded in the
 /*	order as specified, and multiple instances of the same type
 /*	are allowed.
-/* .IP "MAIL_SERVER_RAW_TABLE (CONFIG_RAW_TABLE *)"
+/* .IP "CA_MAIL_SERVER_TIME_TABLE(CONFIG_TIME_TABLE *)"
+/*	A table with configurable parameters, to be loaded from the
+/*	global Postfix configuration file. Tables are loaded in the
+/*	order as specified, and multiple instances of the same type
+/*	are allowed.
+/* .IP "CA_MAIL_SERVER_RAW_TABLE(CONFIG_RAW_TABLE *)"
 /*	A table with configurable parameters, to be loaded from the
 /*	global Postfix configuration file. Tables are loaded in the
 /*	order as specified, and multiple instances of the same type
 /*	are allowed. Raw parameters are not subjected to $name
 /*	evaluation.
-/* .IP "MAIL_SERVER_PRE_INIT (void *(char *service_name, char **argv))"
+/* .IP "CA_MAIL_SERVER_NINT_TABLE(CONFIG_NINT_TABLE *)"
+/*	A table with configurable parameters, to be loaded from the
+/*	global Postfix configuration file. Tables are loaded in the
+/*	order as specified, and multiple instances of the same type
+/*	are allowed.
+/* .IP "CA_MAIL_SERVER_NBOOL_TABLE(CONFIG_NBOOL_TABLE *)"
+/*	A table with configurable parameters, to be loaded from the
+/*	global Postfix configuration file. Tables are loaded in the
+/*	order as specified, and multiple instances of the same type
+/*	are allowed.
+/* .IP "CA_MAIL_SERVER_PRE_INIT(void *(char *service_name, char **argv))"
 /*	A pointer to a function that is called once
 /*	by the skeleton after it has read the global configuration file
 /*	and after it has processed command-line arguments, but before
 /*	the skeleton has optionally relinquished the process privileges.
 /* .sp
 /*	Only the last instance of this parameter type is remembered.
-/* .IP "MAIL_SERVER_POST_INIT (void *(char *service_name, char **argv))"
+/* .IP "CA_MAIL_SERVER_POST_INIT(void *(char *service_name, char **argv))"
 /*	A pointer to a function that is called once
 /*	by the skeleton after it has optionally relinquished the process
 /*	privileges, but before servicing client connection requests.
 /* .sp
 /*	Only the last instance of this parameter type is remembered.
-/* .IP "MAIL_SERVER_LOOP (int *(char *service_name, char **argv))"
+/* .IP "CA_MAIL_SERVER_LOOP(int *(char *service_name, char **argv))"
 /*	A pointer to function that is executed from
 /*	within the event loop, whenever an I/O or timer event has happened,
 /*	or whenever nothing has happened for a specified amount of time.
@@ -90,34 +107,38 @@
 /*	the next event. Specify -1 to wait for "as long as it takes".
 /* .sp
 /*	Only the last instance of this parameter type is remembered.
-/* .IP "MAIL_SERVER_EXIT (void *(char *service_name, char **argv))"
+/* .IP "CA_MAIL_SERVER_EXIT(void *(char *service_name, char **argv))"
 /*	A pointer to function that is executed immediately before normal
 /*	process termination.
-/* .IP "MAIL_SERVER_PRE_ACCEPT (void *(char *service_name, char **argv))"
+/* .IP "CA_MAIL_SERVER_PRE_ACCEPT(void *(char *service_name, char **argv))"
 /*	Function to be executed prior to accepting a new connection.
 /* .sp
 /*	Only the last instance of this parameter type is remembered.
-/* .IP "MAIL_SERVER_PRE_DISCONN (VSTREAM *, void *(char *service_name, char **argv))"
+/* .IP "CA_MAIL_SERVER_PRE_DISCONN(VSTREAM *, char *service_name, char **argv)"
 /*	A pointer to a function that is called
 /*	by the multi_server_disconnect() function (see below).
 /* .sp
 /*	Only the last instance of this parameter type is remembered.
-/* .IP "MAIL_SERVER_IN_FLOW_DELAY (none)"
+/* .IP CA_MAIL_SERVER_IN_FLOW_DELAY
 /*	Pause $in_flow_delay seconds when no "mail flow control token"
 /*	is available. A token is consumed for each connection request.
-/* .IP MAIL_SERVER_SOLITARY
+/* .IP CA_MAIL_SERVER_SOLITARY
 /*	This service must be configured with process limit of 1.
-/* .IP MAIL_SERVER_UNLIMITED
+/* .IP CA_MAIL_SERVER_UNLIMITED
 /*	This service must be configured with process limit of 0.
-/* .IP MAIL_SERVER_PRIVILEGED
+/* .IP CA_MAIL_SERVER_PRIVILEGED
 /*	This service must be configured as privileged.
+/* .IP "CA_MAIL_SERVER_BOUNCE_INIT(const char *, const char **)"
+/*	Initialize the DSN filter for the bounce/defer service
+/*	clients with the specified map source and map names.
 /* .PP
 /*	multi_server_disconnect() should be called by the application
-/*	when a client disconnects.
+/*	to close a client connection.
 /*
 /*	multi_server_drain() should be called when the application
 /*	no longer wishes to accept new client connections. Existing
-/*	clients are handled in a background process. A non-zero
+/*	clients are handled in a background process, and the process
+/*	terminates when the last client is disconnected. A non-zero
 /*	result means this call should be tried again later.
 /*
 /*	The var_use_limit variable limits the number of clients that
@@ -143,15 +164,20 @@
 /*	IBM T.J. Watson Research
 /*	P.O. Box 704
 /*	Yorktown Heights, NY 10598, USA
+/*
+/*	Wietse Venema
+/*	Google, Inc.
+/*	111 8th Avenue
+/*	New York, NY 10011, USA
 /*--*/
 
 /* System library. */
 
 #include <sys_defs.h>
 #include <sys/socket.h>
+#include <sys/time.h>			/* select() */
 #include <unistd.h>
 #include <signal.h>
-#include <syslog.h>
 #include <stdlib.h>
 #include <limits.h>
 #include <string.h>
@@ -163,10 +189,14 @@
 #endif
 #include <time.h>
 
+#ifdef USE_SYS_SELECT_H
+#include <sys/select.h>			/* select() */
+#endif
+
 /* Utility library. */
 
 #include <msg.h>
-#include <msg_syslog.h>
+#include <msg_vstream.h>
 #include <chroot_uid.h>
 #include <listen.h>
 #include <events.h>
@@ -193,6 +223,9 @@
 #include <timed_ipc.h>
 #include <resolve_local.h>
 #include <mail_flow.h>
+#include <mail_version.h>
+#include <bounce.h>
+#include <maillog_client.h>
 
 /* Process manager. */
 
@@ -212,13 +245,14 @@ static int socket_count = 1;
 static void (*multi_server_service) (VSTREAM *, char *, char **);
 static char *multi_server_name;
 static char **multi_server_argv;
-static void (*multi_server_accept) (int, char *);
+static void (*multi_server_accept) (int, void *);
 static void (*multi_server_onexit) (char *, char **);
 static void (*multi_server_pre_accept) (char *, char **);
 static VSTREAM *multi_server_lock;
 static int multi_server_in_flow_delay;
 static unsigned multi_server_generation;
 static void (*multi_server_pre_disconn) (VSTREAM *, char *, char **);
+static int multi_server_saved_flags;
 
 /* multi_server_exit - normal termination */
 
@@ -231,7 +265,7 @@ static NORETURN multi_server_exit(void)
 
 /* multi_server_abort - terminate after abnormal master exit */
 
-static void multi_server_abort(int unused_event, char *unused_context)
+static void multi_server_abort(int unused_event, void *unused_context)
 {
     if (msg_verbose)
 	msg_info("master disconnect -- exiting");
@@ -240,17 +274,18 @@ static void multi_server_abort(int unused_event, char *unused_context)
 
 /* multi_server_timeout - idle time exceeded */
 
-static void multi_server_timeout(int unused_event, char *unused_context)
+static void multi_server_timeout(int unused_event, void *unused_context)
 {
     if (msg_verbose)
 	msg_info("idle timeout -- exiting");
     multi_server_exit();
 }
 
-/*  multi_server_drain - stop accepting new clients */
+/* multi_server_drain - stop accepting new clients */
 
-int multi_server_drain(void)
+int     multi_server_drain(void)
 {
+    const char *myname = "multi_server_drain";
     int     fd;
 
     switch (fork()) {
@@ -259,8 +294,15 @@ int multi_server_drain(void)
 	return (-1);
 	/* Finish existing clients in the background, then terminate. */
     case 0:
-	for (fd = MASTER_LISTEN_FD; fd < MASTER_LISTEN_FD + socket_count; fd++)
+	(void) msg_cleanup((MSG_CLEANUP_FN) 0);
+	event_fork();
+	for (fd = MASTER_LISTEN_FD; fd < MASTER_LISTEN_FD + socket_count; fd++) {
 	    event_disable_readwrite(fd);
+	    (void) close(fd);
+	    /* Play safe - don't reuse this file number. */
+	    if (DUP2(STDIN_FILENO, fd) < 0)
+		msg_warn("%s: dup2(%d, %d): %m", myname, STDIN_FILENO, fd);
+	}
 	var_use_limit = 1;
 	return (0);
 	/* Let the master start a new process. */
@@ -280,14 +322,20 @@ void    multi_server_disconnect(VSTREAM *stream)
     event_disable_readwrite(vstream_fileno(stream));
     (void) vstream_fclose(stream);
     client_count--;
-    use_count++;
+    /* Avoid integer wrap-around in a persistent process.  */
+    if (use_count < INT_MAX)
+	use_count++;
+    if (client_count == 0 && var_idle_limit > 0)
+	event_request_timer(multi_server_timeout, (void *) 0, var_idle_limit);
 }
 
 /* multi_server_execute - in case (char *) != (struct *) */
 
-static void multi_server_execute(int unused_event, char *context)
+static void multi_server_execute(int unused_event, void *context)
 {
     VSTREAM *stream = (VSTREAM *) context;
+    HTABLE *attr = (vstream_flags(stream) == multi_server_saved_flags ?
+		    (HTABLE *) vstream_context(stream) : 0);
 
     if (multi_server_lock != 0
 	&& myflock(vstream_fileno(multi_server_lock), INTERNAL_LOCK,
@@ -295,37 +343,56 @@ static void multi_server_execute(int unused_event, char *context)
 	msg_fatal("select unlock: %m");
 
     /*
-     * Do not bother the application when the client disconnected.
+     * Do not bother the application when the client disconnected. Don't drop
+     * the already accepted client request after "postfix reload"; that would
+     * be rude.
      */
     if (peekfd(vstream_fileno(stream)) > 0) {
 	if (master_notify(var_pid, multi_server_generation, MASTER_STAT_TAKEN) < 0)
-	    multi_server_abort(EVENT_NULL_TYPE, EVENT_NULL_CONTEXT);
+	     /* void */ ;
 	multi_server_service(stream, multi_server_name, multi_server_argv);
 	if (master_notify(var_pid, multi_server_generation, MASTER_STAT_AVAIL) < 0)
 	    multi_server_abort(EVENT_NULL_TYPE, EVENT_NULL_CONTEXT);
     } else {
 	multi_server_disconnect(stream);
     }
-    if (client_count == 0 && var_idle_limit > 0)
-	event_request_timer(multi_server_timeout, (char *) 0, var_idle_limit);
+    if (attr)
+	htable_free(attr, myfree);
 }
 
 /* multi_server_enable_read - enable read events */
 
-static void multi_server_enable_read(int unused_event, char *context)
+static void multi_server_enable_read(int unused_event, void *context)
 {
     VSTREAM *stream = (VSTREAM *) context;
 
-    event_enable_read(vstream_fileno(stream), multi_server_execute, (char *) stream);
+    event_enable_read(vstream_fileno(stream), multi_server_execute, (void *) stream);
 }
 
 /* multi_server_wakeup - wake up application */
 
-static void multi_server_wakeup(int fd)
+static void multi_server_wakeup(int fd, HTABLE *attr)
 {
     VSTREAM *stream;
     char   *tmp;
 
+#if defined(F_DUPFD) && (EVENTS_STYLE != EVENTS_STYLE_SELECT)
+#ifndef THRESHOLD_FD_WORKAROUND
+#define THRESHOLD_FD_WORKAROUND 128
+#endif
+    int     new_fd;
+
+    /*
+     * Leave some handles < FD_SETSIZE for DBMS libraries, in the unlikely
+     * case of a multi-server with a thousand clients.
+     */
+    if (fd < THRESHOLD_FD_WORKAROUND) {
+	if ((new_fd = fcntl(fd, F_DUPFD, THRESHOLD_FD_WORKAROUND)) < 0)
+	    msg_fatal("fcntl F_DUPFD: %m");
+	(void) close(fd);
+	fd = new_fd;
+    }
+#endif
     if (msg_verbose)
 	msg_info("connection established fd %d", fd);
     non_blocking(fd, BLOCKING);
@@ -333,21 +400,25 @@ static void multi_server_wakeup(int fd)
     client_count++;
     stream = vstream_fdopen(fd, O_RDWR);
     tmp = concatenate(multi_server_name, " socket", (char *) 0);
-    vstream_control(stream, VSTREAM_CTL_PATH, tmp, VSTREAM_CTL_END);
+    vstream_control(stream,
+		    CA_VSTREAM_CTL_PATH(tmp),
+		    CA_VSTREAM_CTL_CONTEXT((void *) attr),
+		    CA_VSTREAM_CTL_END);
     myfree(tmp);
     timed_ipc_setup(stream);
+    multi_server_saved_flags = vstream_flags(stream);
     if (multi_server_in_flow_delay && mail_flow_get(1) < 0)
-	event_request_timer(multi_server_enable_read, (char *) stream,
+	event_request_timer(multi_server_enable_read, (void *) stream,
 			    var_in_flow_delay);
     else
-	multi_server_enable_read(0, (char *) stream);
+	multi_server_enable_read(0, (void *) stream);
 }
 
 /* multi_server_accept_local - accept client connection request */
 
-static void multi_server_accept_local(int unused_event, char *context)
+static void multi_server_accept_local(int unused_event, void *context)
 {
-    int     listen_fd = CAST_CHAR_PTR_TO_INT(context);
+    int     listen_fd = CAST_ANY_PTR_TO_INT(context);
     int     time_left = -1;
     int     fd;
 
@@ -359,7 +430,7 @@ static void multi_server_accept_local(int unused_event, char *context)
      * minimize confusion.
      */
     if (client_count == 0 && var_idle_limit > 0)
-	time_left = event_cancel_timer(multi_server_timeout, (char *) 0);
+	time_left = event_cancel_timer(multi_server_timeout, (void *) 0);
 
     if (multi_server_pre_accept)
 	multi_server_pre_accept(multi_server_name, multi_server_argv);
@@ -370,23 +441,24 @@ static void multi_server_accept_local(int unused_event, char *context)
 	msg_fatal("select unlock: %m");
     if (fd < 0) {
 	if (errno != EAGAIN)
-	    msg_fatal("accept connection: %m");
+	    msg_error("accept connection: %m");
 	if (time_left >= 0)
-	    event_request_timer(multi_server_timeout, (char *) 0, time_left);
+	    event_request_timer(multi_server_timeout, (void *) 0, time_left);
 	return;
     }
-    multi_server_wakeup(fd);
+    multi_server_wakeup(fd, (HTABLE *) 0);
 }
 
 #ifdef MASTER_XPORT_NAME_PASS
 
 /* multi_server_accept_pass - accept descriptor */
 
-static void multi_server_accept_pass(int unused_event, char *context)
+static void multi_server_accept_pass(int unused_event, void *context)
 {
-    int     listen_fd = CAST_CHAR_PTR_TO_INT(context);
+    int     listen_fd = CAST_ANY_PTR_TO_INT(context);
     int     time_left = -1;
     int     fd;
+    HTABLE *attr = 0;
 
     /*
      * Be prepared for accept() to fail because some other process already
@@ -396,32 +468,32 @@ static void multi_server_accept_pass(int unused_event, char *context)
      * minimize confusion.
      */
     if (client_count == 0 && var_idle_limit > 0)
-	time_left = event_cancel_timer(multi_server_timeout, (char *) 0);
+	time_left = event_cancel_timer(multi_server_timeout, (void *) 0);
 
     if (multi_server_pre_accept)
 	multi_server_pre_accept(multi_server_name, multi_server_argv);
-    fd = PASS_ACCEPT(listen_fd);
+    fd = pass_accept_attr(listen_fd, &attr);
     if (multi_server_lock != 0
 	&& myflock(vstream_fileno(multi_server_lock), INTERNAL_LOCK,
 		   MYFLOCK_OP_NONE) < 0)
 	msg_fatal("select unlock: %m");
     if (fd < 0) {
 	if (errno != EAGAIN)
-	    msg_fatal("accept connection: %m");
+	    msg_error("accept connection: %m");
 	if (time_left >= 0)
-	    event_request_timer(multi_server_timeout, (char *) 0, time_left);
+	    event_request_timer(multi_server_timeout, (void *) 0, time_left);
 	return;
     }
-    multi_server_wakeup(fd);
+    multi_server_wakeup(fd, attr);
 }
 
 #endif
 
 /* multi_server_accept_inet - accept client connection request */
 
-static void multi_server_accept_inet(int unused_event, char *context)
+static void multi_server_accept_inet(int unused_event, void *context)
 {
-    int     listen_fd = CAST_CHAR_PTR_TO_INT(context);
+    int     listen_fd = CAST_ANY_PTR_TO_INT(context);
     int     time_left = -1;
     int     fd;
 
@@ -433,7 +505,7 @@ static void multi_server_accept_inet(int unused_event, char *context)
      * minimize confusion.
      */
     if (client_count == 0 && var_idle_limit > 0)
-	time_left = event_cancel_timer(multi_server_timeout, (char *) 0);
+	time_left = event_cancel_timer(multi_server_timeout, (void *) 0);
 
     if (multi_server_pre_accept)
 	multi_server_pre_accept(multi_server_name, multi_server_argv);
@@ -444,19 +516,19 @@ static void multi_server_accept_inet(int unused_event, char *context)
 	msg_fatal("select unlock: %m");
     if (fd < 0) {
 	if (errno != EAGAIN)
-	    msg_fatal("accept connection: %m");
+	    msg_error("accept connection: %m");
 	if (time_left >= 0)
-	    event_request_timer(multi_server_timeout, (char *) 0, time_left);
+	    event_request_timer(multi_server_timeout, (void *) 0, time_left);
 	return;
     }
-    multi_server_wakeup(fd);
+    multi_server_wakeup(fd, (HTABLE *) 0);
 }
 
 /* multi_server_main - the real main program */
 
 NORETURN multi_server_main(int argc, char **argv, MULTI_SERVER_FN service,...)
 {
-    char   *myname = "multi_server_main";
+    const char *myname = "multi_server_main";
     VSTREAM *stream = 0;
     char   *root_dir = 0;
     char   *user_name = 0;
@@ -472,15 +544,23 @@ NORETURN multi_server_main(int argc, char **argv, MULTI_SERVER_FN service,...)
     MAIL_SERVER_LOOP_FN loop = 0;
     int     key;
     char   *transport = 0;
+
 #if 0
     char   *lock_path;
     VSTRING *why;
+
 #endif
     int     alone = 0;
     int     zerolimit = 0;
     WATCHDOG *watchdog;
+    char   *oname_val;
+    char   *oname;
     char   *oval;
+    const char *err;
     char   *generation;
+    int     msg_vstream_needed = 0;
+    const char *dsn_filter_title;
+    const char **dsn_filter_maps;
 
     /*
      * Process environment options as early as we can.
@@ -512,10 +592,15 @@ NORETURN multi_server_main(int argc, char **argv, MULTI_SERVER_FN service,...)
      * Initialize logging and exit handler. Do the syslog first, so that its
      * initialization completes before we enter the optional chroot jail.
      */
-    msg_syslog_init(mail_task(var_procname), LOG_PID, LOG_FACILITY);
+    maillog_client_init(mail_task(var_procname), MAILLOG_CLIENT_FLAG_NONE);
     if (msg_verbose)
 	msg_info("daemon started");
 
+    /*
+     * Check the Postfix library version as soon as we enable logging.
+     */
+    MAIL_VERSION_CHECK;
+
     /*
      * Initialize from the configuration file. Allow command-line options to
      * override compiled-in defaults or configured parameter values.
@@ -523,16 +608,17 @@ NORETURN multi_server_main(int argc, char **argv, MULTI_SERVER_FN service,...)
     mail_conf_suck();
 
     /*
-     * Register dictionaries that use higher-level interfaces and protocols.
+     * After database open error, continue execution with reduced
+     * functionality.
      */
-    mail_dict_init();
+    dict_allow_surrogate = 1;
 
     /*
      * Pick up policy settings from master process. Shut up error messages to
      * stderr, because no-one is going to see them.
      */
     opterr = 0;
-    while ((c = GETOPT(argc, argv, "cdDi:lm:n:o:s:St:uvz")) > 0) {
+    while ((c = GETOPT(argc, argv, "cdDi:lm:n:o:s:St:uvVz")) > 0) {
 	switch (c) {
 	case 'c':
 	    root_dir = "setme";
@@ -556,9 +642,11 @@ NORETURN multi_server_main(int argc, char **argv, MULTI_SERVER_FN service,...)
 	    service_name = optarg;
 	    break;
 	case 'o':
-	    if ((oval = split_at(optarg, '=')) == 0)
-		oval = "";
-	    mail_conf_update(optarg, oval);
+	    oname_val = mystrdup(optarg);
+	    if ((err = split_nameval(oname_val, &oname, &oval)) != 0)
+		msg_fatal("invalid \"-o %s\" option value: %s", optarg, err);
+	    mail_conf_update(oname, oval);
+	    myfree(oname_val);
 	    break;
 	case 's':
 	    if ((socket_count = atoi(optarg)) <= 0)
@@ -576,19 +664,40 @@ NORETURN multi_server_main(int argc, char **argv, MULTI_SERVER_FN service,...)
 	case 'v':
 	    msg_verbose++;
 	    break;
+	case 'V':
+	    if (++msg_vstream_needed == 1)
+		msg_vstream_init(mail_task(var_procname), VSTREAM_ERR);
+	    break;
 	case 'z':
 	    zerolimit = 1;
 	    break;
 	default:
-	    msg_fatal("invalid option: %c", c);
+	    msg_fatal("invalid option: %c", optopt);
 	    break;
 	}
     }
+    set_mail_conf_str(VAR_SERVNAME, service_name);
 
     /*
-     * Initialize generic parameters.
+     * Initialize generic parameters and re-initialize logging in case of a
+     * non-default program name or logging destination.
      */
     mail_params_init();
+    maillog_client_init(mail_task(var_procname), MAILLOG_CLIENT_FLAG_NONE);
+
+    /*
+     * Register higher-level dictionaries and initialize the support for
+     * dynamically-loaded dictionarles.
+     */
+    mail_dict_init();
+
+    /*
+     * If not connected to stdin, stdin must not be a terminal.
+     */
+    if (daemon_mode && stream == 0 && isatty(STDIN_FILENO)) {
+	msg_vstream_init(var_procname, VSTREAM_ERR);
+	msg_fatal("do not run this command by hand");
+    }
 
     /*
      * Application-specific initialization.
@@ -599,6 +708,9 @@ NORETURN multi_server_main(int argc, char **argv, MULTI_SERVER_FN service,...)
 	case MAIL_SERVER_INT_TABLE:
 	    get_mail_conf_int_table(va_arg(ap, CONFIG_INT_TABLE *));
 	    break;
+	case MAIL_SERVER_LONG_TABLE:
+	    get_mail_conf_long_table(va_arg(ap, CONFIG_LONG_TABLE *));
+	    break;
 	case MAIL_SERVER_STR_TABLE:
 	    get_mail_conf_str_table(va_arg(ap, CONFIG_STR_TABLE *));
 	    break;
@@ -611,6 +723,12 @@ NORETURN multi_server_main(int argc, char **argv, MULTI_SERVER_FN service,...)
 	case MAIL_SERVER_RAW_TABLE:
 	    get_mail_conf_raw_table(va_arg(ap, CONFIG_RAW_TABLE *));
 	    break;
+	case MAIL_SERVER_NINT_TABLE:
+	    get_mail_conf_nint_table(va_arg(ap, CONFIG_NINT_TABLE *));
+	    break;
+	case MAIL_SERVER_NBOOL_TABLE:
+	    get_mail_conf_nbool_table(va_arg(ap, CONFIG_NBOOL_TABLE *));
+	    break;
 	case MAIL_SERVER_PRE_INIT:
 	    pre_init = va_arg(ap, MAIL_SERVER_INIT_FN);
 	    break;
@@ -647,6 +765,11 @@ NORETURN multi_server_main(int argc, char **argv, MULTI_SERVER_FN service,...)
 		msg_fatal("service %s requires privileged operation",
 			  service_name);
 	    break;
+	case MAIL_SERVER_BOUNCE_INIT:
+	    dsn_filter_title = va_arg(ap, const char *);
+	    dsn_filter_maps = va_arg(ap, const char **);
+	    bounce_client_init(dsn_filter_title, *dsn_filter_maps);
+	    break;
 	default:
 	    msg_panic("%s: unknown argument type: %d", myname, key);
 	}
@@ -658,14 +781,6 @@ NORETURN multi_server_main(int argc, char **argv, MULTI_SERVER_FN service,...)
     if (user_name)
 	user_name = var_mail_owner;
 
-    /*
-     * If not connected to stdin, stdin must not be a terminal.
-     */
-    if (daemon_mode && stream == 0 && isatty(STDIN_FILENO)) {
-	msg_vstream_init(var_procname, VSTREAM_ERR);
-	msg_fatal("do not run this command by hand");
-    }
-
     /*
      * Can options be required?
      */
@@ -746,9 +861,7 @@ NORETURN multi_server_main(int argc, char **argv, MULTI_SERVER_FN service,...)
      * Optionally, restrict the damage that this process can do.
      */
     resolve_local_init();
-#ifdef SNAPSHOT
     tzset();
-#endif
     chroot_uid(root_dir, user_name);
 
     /*
@@ -764,9 +877,9 @@ NORETURN multi_server_main(int argc, char **argv, MULTI_SERVER_FN service,...)
      */
     if (stream != 0) {
 	vstream_control(stream,
-			VSTREAM_CTL_DOUBLE,
-			VSTREAM_CTL_WRITE_FD, STDOUT_FILENO,
-			VSTREAM_CTL_END);
+			CA_VSTREAM_CTL_DOUBLE,
+			CA_VSTREAM_CTL_WRITE_FD(STDOUT_FILENO),
+			CA_VSTREAM_CTL_END);
 	service(stream, multi_server_name, multi_server_argv);
 	vstream_fflush(stream);
 	multi_server_exit();
@@ -779,16 +892,16 @@ NORETURN multi_server_main(int argc, char **argv, MULTI_SERVER_FN service,...)
      * when the master process terminated abnormally.
      */
     if (var_idle_limit > 0)
-	event_request_timer(multi_server_timeout, (char *) 0, var_idle_limit);
+	event_request_timer(multi_server_timeout, (void *) 0, var_idle_limit);
     for (fd = MASTER_LISTEN_FD; fd < MASTER_LISTEN_FD + socket_count; fd++) {
-	event_enable_read(fd, multi_server_accept, CAST_INT_TO_CHAR_PTR(fd));
+	event_enable_read(fd, multi_server_accept, CAST_INT_TO_VOID_PTR(fd));
 	close_on_exec(fd, CLOSE_ON_EXEC);
     }
-    event_enable_read(MASTER_STATUS_FD, multi_server_abort, (char *) 0);
+    event_enable_read(MASTER_STATUS_FD, multi_server_abort, (void *) 0);
     close_on_exec(MASTER_STATUS_FD, CLOSE_ON_EXEC);
     close_on_exec(MASTER_FLOW_READ, CLOSE_ON_EXEC);
     close_on_exec(MASTER_FLOW_WRITE, CLOSE_ON_EXEC);
-    watchdog = watchdog_create(var_daemon_timeout, (WATCHDOG_FN) 0, (char *) 0);
+    watchdog = watchdog_create(var_daemon_timeout, (WATCHDOG_FN) 0, (void *) 0);
 
     /*
      * The event loop, at last.
