@@ -30,8 +30,19 @@
 /*	List the mail queue. Each entry shows the queue file ID, message
 /*	size, arrival time, sender, and the recipients that still need to
 /*	be delivered.  If mail could not be delivered upon the last attempt,
-/*	the reason for failure is shown. This mode of operation is implemented
-/*	by executing the \fBpostqueue\fR(1) command.
+/*	the reason for failure is shown. The queue ID string is
+/*	followed by an optional status character:
+/* .RS
+/* .IP \fB*\fR
+/*	The message is in the \fBactive\fR queue, i.e. the message is
+/*	selected for delivery.
+/* .IP \fB!\fR
+/*	The message is in the \fBhold\fR queue, i.e. no further delivery
+/*	attempt will be made until the mail is taken off hold.
+/* .RE
+/* .IP
+/*	This mode of operation is implemented by executing the
+/*	\fBpostqueue\fR(1) command.
 /* .IP \fBnewaliases\fR
 /*	Initialize the alias database.  If no input file is specified (with
 /*	the \fB-oA\fR option, see below), the program processes the file(s)
@@ -65,6 +76,11 @@
 /* .IP \fB-bi\fR
 /*	Initialize alias database. See the \fBnewaliases\fR
 /*	command above.
+/* .IP \fB-bl\fR
+/*	Go into daemon mode. To accept only local connections as
+/*	with Sendmail\'s \fB-bl\fR option, specify "\fBinet_interfaces
+/*	= loopback\fR" in the Postfix \fBmain.cf\fR configuration
+/*	file.
 /* .IP \fB-bm\fR
 /*	Read mail from standard input and arrange for delivery.
 /*	This is the default mode of operation.
@@ -91,11 +107,17 @@
 /*	parent directory. This information is ignored with Postfix
 /*	versions before 2.3.
 /*
-/*	With older Postfix versions, specify a directory pathname
+/*	With Postfix version 3.2 and later, a non-default directory
+/*	must be authorized in the default \fBmain.cf\fR file, through
+/*	the alternate_config_directories or multi_instance_directories
+/*	parameters.
+/*
+/*	With all Postfix versions, you can specify a directory pathname
 /*	with the MAIL_CONFIG environment variable to override the
 /*	location of configuration files.
-/* .IP "\fB-F \fIfull_name\fR
-/*	Set the sender full name. This is used only with messages that
+/* .IP "\fB-F \fIfull_name\fR"
+/*	Set the sender full name. This overrides the NAME environment
+/*	variable, and is used only with messages that
 /*	have no \fBFrom:\fR message header.
 /* .IP "\fB-f \fIsender\fR"
 /*	Set the envelope sender address. This is the address where
@@ -138,7 +160,8 @@
 /*	\fItype\fR:\fIpathname\fR. See \fBpostalias\fR(1) for
 /*	details.
 /* .IP "\fB-O \fIoption=value\fR (ignored)"
-/*	Backwards compatibility.
+/*	Set the named \fIoption\fR to \fIvalue\fR. Use the equivalent
+/*	configuration parameter in \fBmain.cf\fR instead.
 /* .IP "\fB-o7\fR (ignored)"
 /* .IP "\fB-o8\fR (ignored)"
 /*	To send 8-bit or binary content, use an appropriate MIME encapsulation
@@ -155,9 +178,16 @@
 /*	Set the envelope sender address. This is the address where
 /*	delivery problems are sent to. With Postfix versions before 2.1, the
 /*	\fBErrors-To:\fR message header overrides the error return address.
-/* .IP "\fB-R \fIreturn_limit\fR (ignored)"
-/*	Limit the size of bounced mail. Use the \fBbounce_size_limit\fR
-/*	configuration parameter instead.
+/* .IP "\fB-R \fIreturn\fR"
+/*	Delivery status notification control.  Specify "hdrs" to
+/*	return only the header when a message bounces, "full" to
+/*	return a full copy (the default behavior).
+/*
+/*	The \fB-R\fR option specifies an upper bound; Postfix will
+/*	return only the header, when a full copy would exceed the
+/*	bounce_size_limit setting.
+/*
+/*	This option is ignored before Postfix version 2.10.
 /* .IP \fB-q\fR
 /*	Attempt to deliver all queued mail. This is implemented by
 /*	executing the \fBpostqueue\fR(1) command.
@@ -167,6 +197,11 @@
 /* .IP "\fB-q\fIinterval\fR (ignored)"
 /*	The interval between queue runs. Use the \fBqueue_run_delay\fR
 /*	configuration parameter instead.
+/* .IP \fB-qI\fIqueueid\fR
+/*	Schedule immediate delivery of mail with the specified queue
+/*	ID.  This option is implemented by executing the
+/*	\fBpostqueue\fR(1) command, and is available with Postfix
+/*	version 2.4 and later.
 /* .IP \fB-qR\fIsite\fR
 /*	Schedule immediate delivery of all mail that is queued for the named
 /*	\fIsite\fR. This option accepts only \fIsite\fR names that are
@@ -216,7 +251,7 @@
 /* .ad
 /* .fi
 /*	By design, this program is not set-user (or group) id. However,
-/*	it must handle data from untrusted users or untrusted machines.
+/*	it must handle data from untrusted, possibly remote, users.
 /*	Thus, the usual precautions need to be taken against malicious
 /*	inputs.
 /* DIAGNOSTICS
@@ -227,11 +262,15 @@
 /* .fi
 /* .IP \fBMAIL_CONFIG\fR
 /*	Directory with Postfix configuration files.
-/* .IP \fBMAIL_VERBOSE\fR
+/* .IP "\fBMAIL_VERBOSE\fR (value does not matter)"
 /*	Enable verbose logging for debugging purposes.
-/* .IP \fBMAIL_DEBUG\fR
+/* .IP "\fBMAIL_DEBUG\fR (value does not matter)"
 /*	Enable debugging with an external command, as specified with the
 /*	\fBdebugger_command\fR configuration parameter.
+/* .IP \fBNAME\fR
+/*	The sender full name. This is used only with messages that
+/*	have no \fBFrom:\fR message header. See also the \fB-F\fR
+/*	option above.
 /* CONFIGURATION PARAMETERS
 /* .ad
 /* .fi
@@ -239,6 +278,13 @@
 /*	this program.
 /*	The text below provides only a parameter summary. See
 /*	\fBpostconf\fR(5) for more details including examples.
+/* COMPATIBILITY CONTROLS
+/* .ad
+/* .fi
+/*	Available with Postfix 2.9 and later:
+/* .IP "\fBsendmail_fix_line_endings (always)\fR"
+/*	Controls how the Postfix sendmail command converts email message
+/*	line endings from <CR><LF> into UNIX format (<LF>).
 /* TROUBLE SHOOTING CONTROLS
 /* .ad
 /* .fi
@@ -278,8 +324,9 @@
 /* .IP "\fBhopcount_limit (50)\fR"
 /*	The maximal number of Received:  message headers that is allowed
 /*	in the primary message headers.
-/* .IP "\fBqueue_run_delay (1000s)\fR"
-/*	The time between deferred queue scans by the queue manager.
+/* .IP "\fBqueue_run_delay (300s)\fR"
+/*	The time between deferred queue scans by the queue manager;
+/*	prior to Postfix 2.4 the default value was 1000s.
 /* FAST FLUSH CONTROLS
 /* .ad
 /* .fi
@@ -316,14 +363,12 @@
 /*	The default database type for use in \fBnewaliases\fR(1), \fBpostalias\fR(1)
 /*	and \fBpostmap\fR(1) commands.
 /* .IP "\fBdelay_warning_time (0h)\fR"
-/*	The time after which the sender receives the message headers of
-/*	mail that is still queued.
-/* .IP "\fBenable_errors_to (no)\fR"
-/*	Report mail delivery errors to the address specified with the
-/*	non-standard Errors-To: message header, instead of the envelope
-/*	sender address (this feature is removed with Postfix 2.2, is
-/*	turned off by default with Postfix 2.1, and is always turned on
-/*	with older Postfix versions).
+/*	The time after which the sender receives a copy of the message
+/*	headers of mail that is still queued.
+/* .IP "\fBimport_environment (see 'postconf -d' output)\fR"
+/*	The list of environment parameters that a privileged Postfix
+/*	process will import from a non-Postfix parent process, or name=value
+/*	environment overrides.
 /* .IP "\fBmail_owner (postfix)\fR"
 /*	The UNIX system account that owns the Postfix queue and most Postfix
 /*	daemon processes.
@@ -335,12 +380,22 @@
 /*	append the specified domain name to incomplete addresses.
 /* .IP "\fBsyslog_facility (mail)\fR"
 /*	The syslog facility of Postfix logging.
-/* .IP "\fBsyslog_name (postfix)\fR"
-/*	The mail system name that is prepended to the process name in syslog
-/*	records, so that "smtpd" becomes, for example, "postfix/smtpd".
-/* .IP "\fBtrigger_timeout (10s)\fR"
-/*	The time limit for sending a trigger to a Postfix daemon (for
-/*	example, the \fBpickup\fR(8) or \fBqmgr\fR(8) daemon).
+/* .IP "\fBsyslog_name (see 'postconf -d' output)\fR"
+/*	A prefix that is prepended to the process name in syslog
+/*	records, so that, for example, "smtpd" becomes "prefix/smtpd".
+/* .PP
+/*	Postfix 3.2 and later:
+/* .IP "\fBalternate_config_directories (empty)\fR"
+/*	A list of non-default Postfix configuration directories that may
+/*	be specified with "-c config_directory" on the command line (in the
+/*	case of \fBsendmail\fR(1), with the "-C" option), or via the MAIL_CONFIG
+/*	environment parameter.
+/* .IP "\fBmulti_instance_directories (empty)\fR"
+/*	An optional list of non-default Postfix configuration directories;
+/*	these directories belong to additional Postfix instances that share
+/*	the Postfix executable files and documentation with the default
+/*	Postfix instance, and that are started, stopped, etc., together
+/*	with the default Postfix instance.
 /* FILES
 /*	/var/spool/postfix, mail queue
 /*	/etc/postfix, configuration files
@@ -356,8 +411,12 @@
 /*	postqueue(1), mail queue control
 /*	syslogd(8), system logging
 /* README_FILES
+/* .ad
+/* .fi
 /*	Use "\fBpostconf readme_directory\fR" or
 /*	"\fBpostconf html_directory\fR" to locate this information.
+/* .na
+/* .nf
 /*	DEBUG_README, Postfix debugging howto
 /*	ETRN_README, Postfix ETRN howto
 /*	VERP_README, Postfix VERP howto
@@ -370,6 +429,11 @@
 /*	IBM T.J. Watson Research
 /*	P.O. Box 704
 /*	Yorktown Heights, NY 10598, USA
+/*
+/*	Wietse Venema
+/*	Google, Inc.
+/*	111 8th Avenue
+/*	New York, NY 10011, USA
 /*--*/
 
 /* System library. */
@@ -382,7 +446,6 @@
 #include <stdlib.h>
 #include <signal.h>
 #include <fcntl.h>
-#include <syslog.h>
 #include <time.h>
 #include <errno.h>
 #include <ctype.h>
@@ -395,7 +458,6 @@
 #include <mymalloc.h>
 #include <vstream.h>
 #include <msg_vstream.h>
-#include <msg_syslog.h>
 #include <vstring_vstream.h>
 #include <username.h>
 #include <fullname.h>
@@ -406,12 +468,17 @@
 #include <set_ugid.h>
 #include <connect.h>
 #include <split_at.h>
+#include <name_code.h>
+#include <warn_stat.h>
+#include <clean_env.h>
+#include <maillog_client.h>
 
 /* Global library. */
 
 #include <mail_queue.h>
 #include <mail_proto.h>
 #include <mail_params.h>
+#include <mail_version.h>
 #include <record.h>
 #include <rec_type.h>
 #include <rec_streamlf.h>
@@ -427,8 +494,10 @@
 #include <deliver_request.h>
 #include <mime_state.h>
 #include <header_opts.h>
+#include <mail_dict.h>
 #include <user_acl.h>
 #include <dsn_mask.h>
+#include <mail_parm_split.h>
 
 /* Application-specific. */
 
@@ -470,12 +539,14 @@ typedef struct SM_STATE {
 } SM_STATE;
 
  /*
-  * Mail submission ACL
+  * Mail submission ACL, line-end fixing.
   */
 char   *var_submit_acl;
+char   *var_sm_fix_eol;
 
-static CONFIG_STR_TABLE str_table[] = {
+static const CONFIG_STR_TABLE str_table[] = {
     VAR_SUBMIT_ACL, DEF_SUBMIT_ACL, &var_submit_acl, 0, 0,
+    VAR_SM_FIX_EOL, DEF_SM_FIX_EOL, &var_sm_fix_eol, 1, 0,
     0,
 };
 
@@ -500,7 +571,7 @@ static void output_text(void *context, int rec_type, const char *buf, ssize_t le
 /* output_header - output one message header */
 
 static void output_header(void *context, int header_class,
-			          HEADER_OPTS *header_info,
+			          const HEADER_OPTS *header_info,
 			          VSTRING *buf, off_t offset)
 {
     SM_STATE *state = (SM_STATE *) context;
@@ -534,7 +605,7 @@ static void output_header(void *context, int header_class,
 	    tok822_internalize(state->temp, tpp[0]->head, TOK822_STR_DEFL);
 	    argv_add(rcpt, STR(state->temp), (char *) 0);
 	}
-	myfree((char *) addr_list);
+	myfree((void *) addr_list);
 	tok822_free_tree(tree);
     }
 
@@ -564,7 +635,7 @@ static void output_header(void *context, int header_class,
 /* enqueue - post one message */
 
 static void enqueue(const int flags, const char *encoding,
-		            const char *dsn_envid, int dsn_notify,
+		         const char *dsn_envid, int dsn_ret, int dsn_notify,
 		            const char *rewrite_context, const char *sender,
 		            const char *full_name, char **recipients)
 {
@@ -579,7 +650,7 @@ static void enqueue(const int flags, const char *encoding,
     TOK822 *tp;
     int     rcpt_count = 0;
     enum {
-	STRIP_CR_DUNNO, STRIP_CR_DO, STRIP_CR_DONT
+	STRIP_CR_DUNNO, STRIP_CR_DO, STRIP_CR_DONT, STRIP_CR_ERROR
     }       strip_cr;
     MAIL_STREAM *handle;
     VSTRING *postdrop_command;
@@ -593,12 +664,19 @@ static void enqueue(const int flags, const char *encoding,
     const char *errstr;
     int     addr_count;
     int     level;
+    static NAME_CODE sm_fix_eol_table[] = {
+	SM_FIX_EOL_ALWAYS, STRIP_CR_DO,
+	SM_FIX_EOL_STRICT, STRIP_CR_DUNNO,
+	SM_FIX_EOL_NEVER, STRIP_CR_DONT,
+	0, STRIP_CR_ERROR,
+    };
 
     /*
      * Access control is enforced in the postdrop command. The code here
      * merely produces a more user-friendly interface.
      */
-    if ((errstr = check_user_acl_byuid(var_submit_acl, uid)) != 0)
+    if ((errstr = check_user_acl_byuid(VAR_SUBMIT_ACL,
+				       var_submit_acl, uid)) != 0)
 	msg_fatal_status(EX_NOPERM,
 	  "User %s(%ld) is not allowed to submit mail", errstr, (long) uid);
 
@@ -672,6 +750,9 @@ static void enqueue(const int flags, const char *encoding,
     if (dsn_envid)
 	rec_fprintf(dst, REC_TYPE_ATTR, "%s=%s",
 		    MAIL_ATTR_DSN_ENVID, dsn_envid);
+    if (dsn_ret)
+	rec_fprintf(dst, REC_TYPE_ATTR, "%s=%d",
+		    MAIL_ATTR_DSN_RET, dsn_ret);
     rec_fprintf(dst, REC_TYPE_ATTR, "%s=%s",
 		MAIL_ATTR_RWR_CONTEXT, rewrite_context);
     if (full_name || (full_name = fullname()) != 0)
@@ -679,7 +760,8 @@ static void enqueue(const int flags, const char *encoding,
     rec_fputs(dst, REC_TYPE_FROM, saved_sender);
     if (verp_delims && *saved_sender == 0)
 	msg_fatal_status(EX_USAGE,
-			 "-V option requires non-null sender address");
+		      "%s(%ld): -V option requires non-null sender address",
+			 saved_sender, (long) uid);
     if (encoding)
 	rec_fprintf(dst, REC_TYPE_ATTR, "%s=%s", MAIL_ATTR_ENCODING, encoding);
     if (DEL_REQ_TRACE_FLAGS(flags))
@@ -727,12 +809,13 @@ static void enqueue(const int flags, const char *encoding,
     rec_fputs(dst, REC_TYPE_MESG, "");
     if (DEL_REQ_TRACE_ONLY(flags) != 0) {
 	if (flags & SM_FLAG_XRCPT)
-	    msg_fatal_status(EX_USAGE, "-t option cannot be used with -bv");
+	    msg_fatal_status(EX_USAGE, "%s(%ld): -t option cannot be used with -bv",
+			     saved_sender, (long) uid);
 	if (*saved_sender)
 	    rec_fprintf(dst, REC_TYPE_NORM, "From: %s", saved_sender);
 	rec_fprintf(dst, REC_TYPE_NORM, "Subject: probe");
 	if (recipients) {
-	    rec_fprintf(dst, REC_TYPE_NORM, "To:");
+	    rec_fprintf(dst, REC_TYPE_CONT, "To:");
 	    for (cpp = recipients; *cpp != 0; cpp++) {
 		rec_fprintf(dst, REC_TYPE_NORM, "	%s%s",
 			    *cpp, cpp[1] ? "," : "");
@@ -765,7 +848,11 @@ static void enqueue(const int flags, const char *encoding,
 	 * Process header/body lines.
 	 */
 	skip_from_ = 1;
-	strip_cr = STRIP_CR_DUNNO;
+	strip_cr = name_code(sm_fix_eol_table, NAME_CODE_FLAG_STRICT_CASE,
+			     var_sm_fix_eol);
+	if (strip_cr == STRIP_CR_ERROR)
+	    msg_fatal_status(EX_USAGE,
+		    "invalid %s value: %s", VAR_SM_FIX_EOL, var_sm_fix_eol);
 	for (prev_type = 0; (type = rec_streamlf_get(VSTREAM_IN, buf, var_line_limit))
 	     != REC_TYPE_EOF; prev_type = type) {
 	    if (strip_cr == STRIP_CR_DUNNO && type == REC_TYPE_NORM) {
@@ -783,7 +870,7 @@ static void enqueue(const int flags, const char *encoding,
 		skip_from_ = 0;
 	    }
 	    if (strip_cr == STRIP_CR_DO && type == REC_TYPE_NORM)
-		if (VSTRING_LEN(buf) > 0 && vstring_end(buf)[-1] == '\r')
+		while (VSTRING_LEN(buf) > 0 && vstring_end(buf)[-1] == '\r')
 		    vstring_truncate(buf, VSTRING_LEN(buf) - 1);
 	    if ((flags & SM_FLAG_AEOF) && prev_type != REC_TYPE_CONT
 		&& VSTRING_LEN(buf) == 1 && *STR(buf) == '.')
@@ -842,9 +929,10 @@ static void enqueue(const int flags, const char *encoding,
     }
     if (rcpt_count == 0)
 	msg_fatal_status(EX_USAGE, (flags & SM_FLAG_XRCPT) ?
-			 "No recipient addresses found in message header" :
-			 "Recipient addresses must be specified on"
-			 " the command line or via the -t option");
+		 "%s(%ld): No recipient addresses found in message header" :
+			 "%s(%ld): Recipient addresses must be specified on"
+			 " the command line or via the -t option",
+			 saved_sender, (long) uid);
 
     /*
      * Identify the end of the queue file.
@@ -883,6 +971,15 @@ static void enqueue(const int flags, const char *encoding,
     myfree(saved_sender);
 }
 
+/* tempfail - sanitize exit status after library run-time error */
+
+static void tempfail(void)
+{
+    exit(EX_TEMPFAIL);
+}
+
+MAIL_VERSION_STAMP_DECLARE;
+
 /* main - the main program */
 
 int     main(int argc, char **argv)
@@ -900,14 +997,22 @@ int     main(int argc, char **argv)
     int     n;
     int     flags = SM_FLAG_DEFAULT;
     char   *site_to_flush = 0;
+    char   *id_to_flush = 0;
     char   *encoding = 0;
     char   *qtime = 0;
     const char *errstr;
     uid_t   uid;
-    char   *rewrite_context = MAIL_ATTR_RWR_LOCAL;
+    const char *rewrite_context = MAIL_ATTR_RWR_LOCAL;
     int     dsn_notify = 0;
+    int     dsn_ret = 0;
     const char *dsn_envid = 0;
     int     saved_optind;
+    ARGV   *import_env;
+
+    /*
+     * Fingerprint executables and core dumps.
+     */
+    MAIL_VERSION_STAMP_ALLOCATE;
 
     /*
      * Be consistent with file permissions.
@@ -943,16 +1048,22 @@ int     main(int argc, char **argv)
 	debug_me = 1;
 
     /*
-     * Initialize. Set up logging, read the global configuration file and
-     * extract configuration information. Set up signal handlers so that we
-     * can clean up incomplete output.
+     * Initialize. Set up logging. Read the global configuration file after
+     * command-line processing. Set up signal handlers so that we can clean
+     * up incomplete output.
      */
     if ((slash = strrchr(argv[0], '/')) != 0 && slash[1])
 	argv[0] = slash + 1;
     msg_vstream_init(argv[0], VSTREAM_ERR);
-    msg_syslog_init(mail_task("sendmail"), LOG_PID, LOG_FACILITY);
+    msg_cleanup(tempfail);
+    maillog_client_init(mail_task("sendmail"), MAILLOG_CLIENT_FLAG_NONE);
     set_mail_conf_str(VAR_PROCNAME, var_procname = mystrdup(argv[0]));
 
+    /*
+     * Check the Postfix library version as soon as we enable logging.
+     */
+    MAIL_VERSION_CHECK;
+
     /*
      * Some sites mistakenly install Postfix sendmail as set-uid root. Drop
      * set-uid privileges only when root, otherwise some systems will not
@@ -982,18 +1093,29 @@ int     main(int argc, char **argv)
 	    break;
 	if (c == 'C') {
 	    VSTRING *buf = vstring_alloc(1);
+	    char   *dir;
 
-	    if (setenv(CONF_ENV_PATH,
-		   strcmp(sane_basename(buf, optarg), MAIN_CONF_FILE) == 0 ?
-		       sane_dirname(buf, optarg) : optarg, 1) < 0)
+	    dir = strcmp(sane_basename(buf, optarg), MAIN_CONF_FILE) == 0 ?
+		sane_dirname(buf, optarg) : optarg;
+	    if (strcmp(dir, DEF_CONFIG_DIR) != 0 && geteuid() != 0)
+		mail_conf_checkdir(dir);
+	    if (setenv(CONF_ENV_PATH, dir, 1) < 0)
 		msg_fatal_status(EX_UNAVAILABLE, "out of memory");
 	    vstring_free(buf);
 	}
     }
     optind = saved_optind;
     mail_conf_read();
+    /* Enforce consistent operation of different Postfix parts.	 */
+    import_env = mail_parm_split(VAR_IMPORT_ENVIRON, var_import_environ);
+    update_env(import_env->argv);
+    argv_free(import_env);
+    /* Re-evaluate mail_task() after reading main.cf. */
+    maillog_client_init(mail_task("sendmail"), MAILLOG_CLIENT_FLAG_NONE);
     get_mail_conf_str_table(str_table);
 
+    mail_dict_init();
+
     if (chdir(var_queue_dir))
 	msg_fatal_status(EX_UNAVAILABLE, "chdir %s: %m", var_queue_dir);
 
@@ -1038,7 +1160,8 @@ int     main(int argc, char **argv)
 	    optind++;
 	    continue;
 	}
-	if (strcmp(argv[OPTIND], "-V") == 0) {
+	if (strcmp(argv[OPTIND], "-V") == 0
+	    && argv[OPTIND + 1] != 0 && strlen(argv[OPTIND + 1]) == 2) {
 	    msg_warn("option -V is deprecated with Postfix 2.3; "
 		     "specify -XV instead");
 	    argv[OPTIND] = "-XV";
@@ -1084,6 +1207,10 @@ int     main(int argc, char **argv)
 	    if ((dsn_notify = dsn_notify_mask(optarg)) == 0)
 		msg_warn("bad -N option value -- ignored");
 	    break;
+	case 'R':
+	    if ((dsn_ret = dsn_ret_code(optarg)) == 0)
+		msg_warn("bad -R option value -- ignored");
+	    break;
 	case 'V':				/* DSN, was: VERP */
 	    if (strlen(optarg) > 100)
 		msg_warn("too long -V option value -- ignored");
@@ -1109,6 +1236,7 @@ int     main(int argc, char **argv)
 	    default:
 		msg_fatal_status(EX_USAGE, "unsupported: -%c%c", c, *optarg);
 	    case 'd':				/* daemon mode */
+	    case 'l':				/* daemon mode */
 		if (mode == SM_MODE_FLUSHQ)
 		    msg_warn("ignoring -q option in daemon mode");
 		mode = SM_MODE_DAEMON;
@@ -1173,6 +1301,10 @@ int     main(int argc, char **argv)
 		site_to_flush = optarg + 1;
 		if (*site_to_flush == 0)
 		    msg_fatal_status(EX_USAGE, "specify: -qRsitename");
+	    } else if (optarg[0] == 'I') {
+		id_to_flush = optarg + 1;
+		if (*id_to_flush == 0)
+		    msg_fatal_status(EX_USAGE, "specify: -qIqueueid");
 	    } else {
 		msg_fatal_status(EX_USAGE, "-q%c is not implemented",
 				 optarg[0]);
@@ -1198,11 +1330,16 @@ int     main(int argc, char **argv)
     if (site_to_flush && mode != SM_MODE_ENQUEUE)
 	msg_fatal_status(EX_USAGE, "-qR can be used only in delivery mode");
 
+    if (id_to_flush && mode != SM_MODE_ENQUEUE)
+	msg_fatal_status(EX_USAGE, "-qI can be used only in delivery mode");
+
     if (flags & DEL_REQ_FLAG_USR_VRFY) {
 	if (flags & SM_FLAG_XRCPT)
 	    msg_fatal_status(EX_USAGE, "-t option cannot be used with -bv");
 	if (dsn_notify)
 	    msg_fatal_status(EX_USAGE, "-N option cannot be used with -bv");
+	if (dsn_ret)
+	    msg_fatal_status(EX_USAGE, "-R option cannot be used with -bv");
 	if (msg_verbose == 1)
 	    msg_fatal_status(EX_USAGE, "-v option cannot be used with -bv");
     }
@@ -1226,20 +1363,32 @@ int     main(int argc, char **argv)
 	msg_panic("unknown operation mode: %d", mode);
 	/* NOTREACHED */
     case SM_MODE_ENQUEUE:
-	if (site_to_flush == 0) {
-	    enqueue(flags, encoding, dsn_envid, dsn_notify,
+	if (site_to_flush) {
+	    if (argv[OPTIND])
+		msg_fatal_status(EX_USAGE, "flush site requires no recipient");
+	    ext_argv = argv_alloc(2);
+	    argv_add(ext_argv, "postqueue", "-s", site_to_flush, (char *) 0);
+	    for (n = 0; n < msg_verbose; n++)
+		argv_add(ext_argv, "-v", (char *) 0);
+	    argv_terminate(ext_argv);
+	    mail_run_replace(var_command_dir, ext_argv->argv);
+	    /* NOTREACHED */
+	} else if (id_to_flush) {
+	    if (argv[OPTIND])
+		msg_fatal_status(EX_USAGE, "flush queue_id requires no recipient");
+	    ext_argv = argv_alloc(2);
+	    argv_add(ext_argv, "postqueue", "-i", id_to_flush, (char *) 0);
+	    for (n = 0; n < msg_verbose; n++)
+		argv_add(ext_argv, "-v", (char *) 0);
+	    argv_terminate(ext_argv);
+	    mail_run_replace(var_command_dir, ext_argv->argv);
+	    /* NOTREACHED */
+	} else {
+	    enqueue(flags, encoding, dsn_envid, dsn_ret, dsn_notify,
 		    rewrite_context, sender, full_name, argv + OPTIND);
 	    exit(0);
+	    /* NOTREACHED */
 	}
-	if (argv[OPTIND])
-	    msg_fatal_status(EX_USAGE, "flush site requires no recipient");
-	ext_argv = argv_alloc(2);
-	argv_add(ext_argv, "postqueue", "-s", site_to_flush, (char *) 0);
-	for (n = 0; n < msg_verbose; n++)
-	    argv_add(ext_argv, "-v", (char *) 0);
-	argv_terminate(ext_argv);
-	mail_run_replace(var_command_dir, ext_argv->argv);
-	/* NOTREACHED */
 	break;
     case SM_MODE_MAILQ:
 	if (argv[OPTIND])
@@ -1286,7 +1435,7 @@ int     main(int argc, char **argv)
 	argv_add(ext_argv, "postalias", (char *) 0);
 	for (n = 0; n < msg_verbose; n++)
 	    argv_add(ext_argv, "-v", (char *) 0);
-	argv_split_append(ext_argv, var_alias_db_map, ", \t\r\n");
+	argv_split_append(ext_argv, var_alias_db_map, CHARS_COMMA_SP);
 	argv_terminate(ext_argv);
 	mail_run_replace(var_command_dir, ext_argv->argv);
 	/* NOTREACHED */
@@ -1295,7 +1444,8 @@ int     main(int argc, char **argv)
 	    msg_fatal_status(EX_USAGE,
 			     "stand-alone mode requires no recipient");
 	/* The actual enforcement happens in the postdrop command. */
-	if ((errstr = check_user_acl_byuid(var_submit_acl, uid = getuid())) != 0)
+	if ((errstr = check_user_acl_byuid(VAR_SUBMIT_ACL, var_submit_acl,
+					   uid = getuid())) != 0)
 	    msg_fatal_status(EX_NOPERM,
 			     "User %s(%ld) is not allowed to submit mail",
 			     errstr, (long) uid);
